@@ -1,3 +1,3 @@
+from . import core, examples, simplified
 from .core import Nx3D
-from .examples import diffusion, game_of_life
 from .simplified import demo, plot
